/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: editor.cc,v 1.13 2000/11/04 16:35:47 pbd Exp $
*/

#include <stdio.h> // for sprintf, grrr 
#include <stdlib.h>
#include <math.h>

#include <string>

#include <gtkwaveform-0.2/gtkedlwavebuffer.h>
#include <gtkwaveform-0.2/gtkwaveview.h>

#include <pbd/error.h>

#include <gtkmmext/utils.h>
#include <gtkmmext/popup_selector.h>

#include "ardour_ui.h"
#include "editor.h"
#include "config.h"
#include "tape.h"
#include "track.h"
#include "filebuffer.h"
#include "constbuffer.h"
#include "utils.h"

using namespace ARDOUR;
using namespace SigC;

const guint32 Editor::max_edit_groups = 32;
static const gchar *region_selector_titles[1] = {
	"Title"
};

GdkCursor *Editor::cross_hair_cursor = 0;

Editor::Editor ()
	: region_xscroll_adjustment (0, 0, 100, 1.0, 10.0),
	  region_xscroller (region_xscroll_adjustment),
	  region_selector (_region_refiller, this, 1, true, true, region_selector_titles),
	  xscroll_adjustment (0, 0, 100, 1.0, 10.0),
	  xscroller (xscroll_adjustment),
	  color_dialog ("ardour: tba"),
	  click_bpm_adjustment (60, 1, 500, 1, 10),
	  click_bpb_adjustment (4, 1, 16, 1, 4),
	  click_note_length_adjustment (4, 1, 16, 1, 4),
	  click_bpm_scale (click_bpm_adjustment),
	  click_note_spinner (click_note_length_adjustment),
	  click_bpb_spinner (click_bpb_adjustment),
	  click_ok_button ("Print click track"),
	  click_cancel_button ("Cancel")
{
	_ok = false;

	tape = 0;
	edits_legal = true;
	escape_as_meta = false;
	prefix_pending = false;
	control_x_pending = false;
	control_c_pending = false;
	scroll_timeout_tag = 0;
	restart_scrolling_when_active = false;
	cut_buffer = 0;
	selected_region = 0;
	click_piece = 0;
	click_emphasis_piece = 0;
	click_emphasis_sound = 0;
	click_sound = 0;

	silent_buffer = const_buffer_new (0);

	fakeTransportDisplay.editor = this;

	build_menu_bar ();

	track_edit_menu = build_track_edit_menu ();
	region_edit_menu = build_track_edit_menu ();

	build_region_window ();

	channel_scroller_base.set_name ("EditorWindow");
	channel_scroller_base.add (view_packer);

	channel_scroller.add_with_viewport (channel_scroller_base);
	channel_scroller.set_policy (GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

	xscroller_box.pack_start (xscroller, true, true);

	cmd_indicator.set_name ("CommandIndicator");
	Gtkmmext::set_usize_to_display_given_text
		(cmd_indicator, "the longest command msg i can imagine", 5, 5);
	cmd_indicator_box.pack_start (cmd_indicator, false, false);

	cmd_indicator_frame.set_shadow_type (GTK_SHADOW_IN);
	cmd_indicator_frame.set_name ("BaseFrame");
	cmd_indicator_frame.add (cmd_indicator_box);

	bottom_box.pack_start (cmd_indicator_frame, false, false, 10);

	vpacker.pack_start (menu_bar, false, false);
	vpacker.pack_start (channel_scroller, true, true);
	vpacker.pack_start (xscroller_box, false, false);
	vpacker.pack_start (bottom_box, false, false, 10);

	hpacker.pack_start (vpacker, true, true);

	set_name ("EditorWindow");
	add (hpacker);
	
	xscroll_adjustment.value_changed.connect (slot (this, &Editor::xscroll_value_changed));

	build_click_window ();

	_ok = true;
}

void
Editor::realize_impl ()

{
	track_edit_menu->accelerate (*this->get_toplevel());
	Gtk::EventBox::realize_impl ();
}

void
Editor::xscroll_value_changed ()

{
	vector<TrackView *>::iterator i;
	gfloat position = xscroll_adjustment.get_value();

	for (i  = visible_track_views.begin();
	     i != visible_track_views.end();
	     i++) {
		(*i)->x_adjuster->set_value (position);
	}
}

void
Editor::region_xscroll_value_changed ()

{
	gfloat position = region_xscroll_adjustment.get_value();
	region_view->x_adjuster->set_value (position);
}

void
Editor::tape_transport_state_change ()

{
	Gtkmmext::UI::instance()->touch_display (&fakeTransportDisplay);
}

void
Editor::start_scrolling ()

{
	if (tape && scroll_timeout_tag <= 0) {
		scroll_timeout_tag = gtk_timeout_add (100, _scroll_timeout, this);
		restart_scrolling_when_active = false;
	}
}

void
Editor::stop_scrolling ()

{
	if (scroll_timeout_tag > 0) {
		gtk_timeout_remove (scroll_timeout_tag);
		restart_scrolling_when_active = false;
		scroll_timeout_tag = -1;
	}
}

void
Editor::map_transport_state ()

{
	if (tape == 0) {
		return;
	}

	edits_legal = false;

	if (tape->transport_state() == Tape::Roll) {
		start_scrolling ();
	} else if (tape->transport_state() == Tape::Stopped) {

		edits_legal = true;

		/* the scroll timeout handler will detect the stopped
		   state, and remove itself.
		*/
	}
}

gint
Editor::_scroll_timeout (void *arg)
{
	return ((Editor *) arg)->scroll_timeout ();
}

gint
Editor::scroll_timeout ()

{
	size_t frame = tape->transport_frame();
	
	reposition_cursor (frame);
		
	if (tape->transport_state() == Tape::Roll) {
		return TRUE;
	} else {
		size_t page = (size_t) track_views[0]->x_adjuster->get_page_size();
		
		page /= 2;
		
		if (frame > page) {
			frame -= page;
		} else {
			frame = 0;
		}
		
		reposition_x_origin (frame);
		
		scroll_timeout_tag = -1;
		return FALSE;
	}
}	

void
Editor::reposition_cursor (size_t sample)

{
	vector<TrackView *>::iterator i;

	for (i  = visible_track_views.begin();
	     i != visible_track_views.end();
	     i++) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW((*i)->waveview), sample);
	}
}

void
Editor::reposition_x_origin (size_t sample)

{
	vector<TrackView *>::iterator i;

	for (i  = visible_track_views.begin();
	     i != visible_track_views.end();
	     i++) {
		(*i)->x_adjuster->set_value (sample);
	}
}

void
Editor::connect_to_tape (Tape *t)

{
	size_t n;
	vector<TrackView *>::iterator iter;

	tape = t;
	tape->going_away.connect (slot (this, &Editor::tape_going_away));
	tape->TransportStateChange.connect (slot (this, &Editor::tape_transport_state_change));

	track_views.reserve (tape->ntracks());

	/* Create however track views many we may need */

	for (n = track_views.size(); n < tape->ntracks(); n++) {
		TrackView *tv;
		
		tv = new TrackView (this, n);
		track_views[n] = tv;

		/* Bind key presses and button releases for the
		   drawing area.
		*/

		tv->wavearea->key_press_event.connect (bind (slot (this, &Editor::key_press_handler), tv));
		tv->wavearea->button_press_event.connect (bind (slot (this, &Editor::button_press_handler), tv));
		tv->wavearea->button_release_event.connect (bind (slot (this, &Editor::button_release_handler), tv));
	}

	/* Hide them all */
	
	for (iter  = visible_track_views.begin();
	     iter != visible_track_views.end();
	     iter++) {
		(*iter)->hide_all ();
		(*iter)->unparent ();
	}

	visible_track_views.clear ();

	/* Now just show the ones we need */

	for (n = 0; n < tape->ntracks(); n++) {
		view_packer.pack_start (*track_views[n], true, true);
		view_packer.pack_start (track_views[n]->sep_box, false, false);

		track_views[n]->reset (tape);
		visible_track_views.push_back (track_views[n]);
		track_views[n]->show_all ();
		track_views[n]->sep_box.show_all ();
	}

//  should work, but shows "hidden" views as blank boxes
//	view_packer.show_all ();

	reset_xscroll_steps (xscroll_adjustment, visible_track_views[0]);
}

void
Editor::reset_xscroll_steps (Gtk::Adjustment &adjustment, Editor::TrackView *view)

{
	Gtk::Adjustment *adj;

	/* Use the adjuster for the first track view. We assume that
	   it has representative values.
	*/

	adj = view->x_adjuster;
	   
	adjustment.set_lower (adj->get_lower());
	adjustment.set_upper (adj->get_upper());
	adjustment.set_step_increment (adj->get_step_increment());
	adjustment.set_page_increment (adj->get_page_increment());
	adjustment.set_page_size (adj->get_page_size());

	adjustment.changed();
}

void
Editor::tape_going_away (AudioThing *)

{
	tape = 0;
}

void
Editor::build_menu_bar ()

{
	using namespace Gtk;
	using namespace Gtk::Menu_Helpers;

	menu_bar.set_name ("EditorMenuBar");

	MenuList& items = menu_bar.items();

	/* file menu */
	
	Menu *file_menu = new Menu();
	MenuList& file_items = file_menu->items();
	file_items.push_back (MenuElem ("Export"));
	file_items.push_back (SeparatorElem());
	file_items.push_back (MenuElem ("Quit"));
	items.push_back (MenuElem ("File", *file_menu));

	/* windows menu */

	Menu *window_menu = new Menu();
	MenuList& window_items = window_menu->items();
	window_items.push_back (MenuElem ("Recorder"));
	window_items.back()->activate.connect
		(bind (slot (ARDOUR_UI::instance(), &ARDOUR_UI::switch_to_notebook_page), 0));

	window_items.push_back (SeparatorElem());

	window_items.push_back 
		(CheckMenuElem ("Transport",
				slot (ARDOUR_UI::instance(), 
				      &ARDOUR_UI::toggle_location_window)));

	window_items.push_back 
		(CheckMenuElem ("Controls",
				slot (ARDOUR_UI::instance(), 
				      &ARDOUR_UI::toggle_control_window)));
	window_items.push_back 
		(CheckMenuElem ("Regions",
				slot (this, &Editor::toggle_region_window)));

	items.push_back (MenuElem ("Windows", *window_menu));
}

void
Editor::set_cross_hair_cursor (Gtk::DrawingArea *area)

{
	if (!cross_hair_cursor) {
		cross_hair_cursor = gdk_cursor_new (GDK_CROSSHAIR);
	}

	gdk_window_set_cursor (area->get_window(), cross_hair_cursor);
}

Gtk::Menu *
Editor::build_track_edit_menu ()

{
	using namespace Gtk::Menu_Helpers;
	Gtk::Menu *the_menu = manage (new Gtk::Menu);
	

	MenuList& edit_items = the_menu->items();

	edit_items.push_back (MenuElem ("Undo", "<control>z", slot (this, &Editor::undo)));
	edit_items.push_back (MenuElem ("Redo", "<control>r", slot (this, &Editor::redo)));

	Gtk::Menu *track_menu = manage (new Gtk::Menu);
	MenuList& track_items = track_menu->items();

	track_items.push_back (MenuElem ("Save track", slot (this, &Editor::save_track)));
	track_items.push_back (MenuElem ("Save track as", slot (this, &Editor::save_track_as)));
	
	Gtk::Menu *export_menu = manage (new Gtk::Menu);
	MenuList& export_items = export_menu->items ();

	export_items.push_back (MenuElem ("Mono WAV file"));
	export_items.push_back (MenuElem ("Stereo WAV file"));
	export_items.push_back (MenuElem ("Mix down to stereo WAV file"));
	export_items.push_back (MenuElem ("Linearized native format"));

	track_items.push_back (MenuElem ("Export", *export_menu));

	edit_items.push_back (MenuElem ("File", *track_menu));

	Gtk::Menu *play_menu = manage (new Gtk::Menu);
	MenuList& play_items = play_menu->items();

	play_items.push_back (MenuElem ("Play from location", " ", slot (this, &Editor::toggle_playback)));
	play_items.push_back (MenuElem ("Play from cursor"));
	play_items.push_back (MenuElem ("Play from start", "<alt> ", slot (this, &Editor::toggle_playback)));
	play_items.push_back (MenuElem ("Play selection"));
	play_items.push_back (MenuElem ("Loop selection"));

	edit_items.push_back (MenuElem ("Play", *play_menu));

	Gtk::Menu *select_menu = manage (new Gtk::Menu);
	MenuList& select_items = select_menu->items();

	select_items.push_back (MenuElem ("Select all"));
	select_items.push_back (MenuElem ("Start kbd. selection", "<control> "));
	select_items.push_back (MenuElem ("Select between marks"));

	edit_items.push_back (MenuElem ("Select", *select_menu));

	Gtk::Menu *cutnpaste_menu = manage (new Gtk::Menu);
	MenuList& cutnpaste_items = cutnpaste_menu->items();

	cutnpaste_items.push_back (MenuElem ("Cut selection", 
					     "<control>w", (bind (slot (this, &Editor::cut_selection), true))));
	cutnpaste_items.push_back (MenuElem ("Copy selection", 
					     "<alt>w", (bind (slot (this, &Editor::cut_selection), false))));
	cutnpaste_items.push_back (MenuElem ("Insert cut buffer", 
					     GDK_Insert, slot (this, &Editor::insert_cut_buffer)));

	edit_items.push_back (MenuElem ("Cut/Paste", *cutnpaste_menu));

	Gtk::Menu *delete_menu = manage (new Gtk::Menu);
	MenuList& delete_items = delete_menu->items();

	delete_items.push_back (MenuElem ("Delete screen", 
					  "<control>k", slot (this, &Editor::delete_screen)));
	delete_items.push_back (MenuElem ("Delete sample", 
					  GDK_Delete, slot (this, &Editor::delete_sample_forward)));
	delete_items.push_back (MenuElem ("Delete sample backwards",
					  "<control>h", slot (this, &Editor::delete_sample_backward)));

	edit_items.push_back (MenuElem ("Delete", *delete_menu));

	Gtk::Menu *silence_menu = manage (new Gtk::Menu);
	MenuList& silence_items = silence_menu->items();

	silence_items.push_back (MenuElem ("Silence selection", 
					   "<control>s", slot (this, &Editor::silence_selection)));
	silence_items.push_back (MenuElem ("Silence sample", 
					   "<control>0", slot (this, &Editor::zero_sample_at_cursor)));
	silence_items.push_back (MenuElem ("Insert silent sample", 
					   "<control>o", slot (this, &Editor::insert_silence)));

	edit_items.push_back (MenuElem ("Silence", *silence_menu));

	Gtk::Menu *zoom_menu = manage (new Gtk::Menu());
	MenuList& zoom_items = zoom_menu->items();

	zoom_items.push_back (MenuElem ("Zoom in", 
					'=', (bind (slot (this, &Editor::temporal_zoom_step), true))));
	zoom_items.push_back (MenuElem ("Zoom out", 
					'-', (bind (slot (this, &Editor::temporal_zoom_step), false))));
	zoom_items.push_back (MenuElem ("Zoom to selection",
					(slot (this, &Editor::temporal_zoom_selection))));
	zoom_items.push_back (MenuElem ("Zoom 1:1",
					(bind (slot (this, &Editor::temporal_zoom), 1.0))));
	zoom_items.push_back (MenuElem ("Show All",
					(slot (this, &Editor::temporal_zoom_tape))));

	edit_items.push_back (MenuElem ("Zoom", *zoom_menu));

	Gtk::Menu *move_menu = manage (new Gtk::Menu);
	MenuList& move_items = move_menu->items ();

	move_items.push_back (MenuElem ("Goto beginning", 
					GDK_Home, slot (this, &Editor::move_to_start)));
	move_items.push_back (MenuElem ("Goto end", 
					GDK_End, slot (this, &Editor::move_to_end)));
	move_items.push_back (MenuElem ("Forward one screen", 
					GDK_Page_Down, slot (this, &Editor::scroll_forward_screen)));
	move_items.push_back (MenuElem ("Back one screen", 
					GDK_Page_Up, slot (this, &Editor::scroll_back_screen)));
	move_items.push_back (MenuElem ("Forward sample", 
					"<control>f", slot (this, &Editor::scroll_forward_sample)));
	move_items.push_back (MenuElem ("Back sample", 
					"<control>b", slot (this, &Editor::scroll_back_sample)));
	move_items.push_back (MenuElem ("Center cursor", 
					"<control>c", slot (this, &Editor::center_cursor)));

	edit_items.push_back (MenuElem ("Motion", *move_menu));

	Gtk::Menu *mark_menu = manage (new Gtk::Menu);
	MenuList& mark_items = mark_menu->items();

	mark_items.push_back (MenuElem ("Set mark at cursor", 
					"<control>m", slot (this, &Editor::set_mark)));
	mark_items.push_back (MenuElem ("Set named mark at cursor", 
					"<alt><control>m", slot (this, &Editor::set_mark)));
	mark_items.push_back (MenuElem ("ALL set mark at cursor", 
					"<><control>m", slot (this, &Editor::set_mark)));
	mark_items.push_back (MenuElem ("Move cursor to next mark", 
					"<control>j", slot (this, &Editor::jump_to_mark)));
	mark_items.push_back (MenuElem ("Move cursor to named mark", 
					"<alt><control>j", slot (this, &Editor::jump_to_mark)));

	edit_items.push_back (MenuElem ("Marks", *mark_menu));

	Gtk::Menu *env_menu = manage (new Gtk::Menu);
	MenuList& env_items = env_menu->items();

	env_items.push_back (MenuElem ("Edit amplitude envelope"));
	env_items.push_back (MenuElem ("Normalize"));
	env_items.push_back (MenuElem ("Scale to ..."));
	
	edit_items.push_back (MenuElem ("Envelopes", *env_menu));

	edit_items.push_back (MenuElem ("Search"));

	Gtk::Menu *color_menu = manage (new Gtk::Menu);
	MenuList& color_items = color_menu->items();
	
	color_items.push_back (MenuElem ("Normal wave", 
					 bind (slot (this, &Editor::set_color), NormalWave)));
	color_items.push_back (MenuElem ("Normal background",
					 bind (slot (this, &Editor::set_color), NormalBackground)));
	color_items.push_back (MenuElem ("Active wave", 
					 bind (slot (this, &Editor::set_color), ActiveWave)));
	color_items.push_back (MenuElem ("Active background",
					 bind (slot (this, &Editor::set_color), ActiveBackground)));
	color_items.push_back (MenuElem ("Selected wave", 
					 bind (slot (this, &Editor::set_color), SelectedWave)));
	color_items.push_back (MenuElem ("Selected background", 
					 bind (slot (this, &Editor::set_color), SelectedBackground)));

	edit_items.push_back (MenuElem ("Colors", *color_menu));

	edit_items.push_back (MenuElem ("Click", slot (this, &Editor::do_click)));

	return the_menu;
}

void
Editor::build_region_window ()

{
	region_window.set_title ("ardour: regions");
	region_window.set_name ("RegionWindow");
	region_window.add (region_hpacker);
	region_window.set_usize (700, 200);

	region_edit_menu->accelerate (region_window);

	region_hpacker.pack_start (region_left_vpacker, true, true, 10);
	region_hpacker.pack_start (region_right_vpacker, true, true, 10);

	region_xscroller_hbox.pack_start (region_xscroller, true, true, 0);
	
	region_view = new TrackView (this, -1, false);

	/* set the edit group so that it never interacts with anything else */

	region_view->edit_group = max_edit_groups + 1;

	region_wave_vbox.pack_start (*region_view, true, true);
	region_wave_vbox.pack_start (region_xscroller_hbox, false, false);

	region_selector.clist().set_column_auto_resize (0, true);
	region_selector.clist().set_column_auto_resize (1, true);
	region_selector.clist().set_column_auto_resize (2, true);
	region_selector.clist().set_name ("RegionSelector");

	region_left_vpacker.pack_start (region_selector, true, true, 10);
	region_right_vpacker.pack_start (region_wave_vbox, true, true, 10);

	region_selector.selection_made.connect (slot (this, &Editor::region_selected));
	region_selector.choice_made.connect (slot (this, &Editor::region_chosen));
	region_xscroll_adjustment.value_changed.connect (slot (this, &Editor::region_xscroll_value_changed));

	region_view->wavearea->key_press_event.connect 
		(bind (slot (this, &Editor::key_press_handler), region_view));
	region_view->wavearea->button_press_event.connect 
		(bind (slot (this, &Editor::button_press_handler), region_view));
	region_view->wavearea->button_release_event.connect 
		(bind (slot (this, &Editor::button_release_handler), region_view));
}

void
Editor::region_selected (Gtkmmext::Selector *sel, Gtkmmext::SelectionResult *res)

{
	list<Region *>::iterator i;
	int n;

	for (n = 0, i = regions.begin(); i != regions.end() && n < res->row; n++, i++);

	if (i != regions.end()) {
		selected_region = *i;
		region_view->set_buffer ((*i)->waveview_buffer);
		reset_xscroll_steps (region_xscroll_adjustment, region_view);
	}
}

void
Editor::region_chosen (Gtkmmext::Selector *sel, Gtkmmext::SelectionResult *res)

{
	Gtk::Dialog dialog;
	Gtk::Entry  entry;
	Gtk::Button ok_button ("OK");
	Gtk::Button cancel_button ("Cancel");

	dialog.set_title ("ardour: rename region");
	dialog.set_name ("RegionRenameWindow");
	dialog.set_usize (300, -1);
	dialog.set_position (GTK_WIN_POS_MOUSE);

	dialog.get_vbox()->set_border_width (10);
	dialog.get_vbox()->pack_start (entry);
	dialog.get_action_area()->pack_start (ok_button);
	dialog.get_action_area()->pack_start (cancel_button);

	entry.set_name ("RegionNameDisplay");
	ok_button.set_name ("EditorGTKButton");
	cancel_button.set_name ("EditorGTKButton");

	region_renamed = false;

	ok_button.clicked.connect (bind (slot (this, &Editor::region_rename_finished), true));
	cancel_button.clicked.connect (bind (slot (this, &Editor::region_rename_finished), false));

	/* recurse */

	dialog.show_all ();
	Gtk::Main::run ();

	if (region_renamed) {
		list<Region *>::iterator i;
		int n;
		for (n = 0, i = regions.begin(); i != regions.end() && n < res->row; n++, i++);
		if (i != regions.end()) {
			(*i)->set_name (entry.get_text ());
		}
		sel->rescan ();
	}
}

void
Editor::region_rename_finished (bool status)

{
	region_renamed = status;
	Gtk::Main::quit ();
}

void
Editor::toggle_region_window ()

{
	if (region_window.is_visible()) {
		region_window.hide_all ();
	} else {
		region_window.show_all ();
	}
}

void
Editor::_region_refiller (Gtk::CList &list, void *arg)
{
	((Editor *) arg)->region_refiller (list);
}

void
Editor::region_refiller (Gtk::CList &clist)
{
	const gchar *rowdata[1];
	int row;
	list<Region *>::iterator i;

	for (row = 0, i = regions.begin(); i != regions.end(); i++) {
		rowdata[0] = (*i)->name.c_str();
		clist.insert_row (row++, rowdata);
	}
	
	if (row) {
		clist.row(0).select();
	}
}

void
Editor::solo_toggle (gint32 tracknum)
{
	tape->set_track_solo (tracknum, track_views[tracknum]->solo_button.get_active());
}

void
Editor::mute_toggle (gint32 tracknum)
{
	tape->set_track_mute (tracknum, track_views[tracknum]->mute_button.get_active());
}

gint
Editor::edit_click (GdkEventButton *ev, gint32 tracknum)

{
	char buf[32];
	TrackView *tv = track_views[tracknum];

	if (ev->button == 2 || (ev->state & GDK_CONTROL_MASK)) {
		tv->edit_group = 0;
	} else if (ev->button == 3) {
		if (tv->edit_group) {
			tv->edit_group--;
		} else {
			tv->edit_group = max_edit_groups - 1;
		}
	} else {
		tv->edit_group++;
		tv->edit_group %= max_edit_groups;
	}

	sprintf (buf, "TrackEditIndicator%d", tv->edit_group);

	tv->edit_button.set_name (buf);
	tv->edit_label.set_name (buf);

	if (tv->edit_group) {
		sprintf (buf, "group %d", tv->edit_group);
	} else {
		sprintf (buf, "group");
	}
	
	tv->edit_label.set_text (buf);

	return FALSE;
}

vector<Editor::TrackView *> *
Editor::get_valid_views (gint32 edit_group)

{
	vector<TrackView *> *v = new vector<TrackView *>;
	vector<TrackView *>::iterator i;

	if (region_view->edit_group == edit_group) {
		v->push_back (region_view);
	}

	for (i  = visible_track_views.begin();
	     i != visible_track_views.end ();
	     i++) {
		if ((*i)->edit_group == edit_group) {
			v->push_back (*i);
		}
	}
	
	return v;
}

/* keymap handling

 */

/* From snd:

  [Down]  zoom out
  [Up]    zoom in
  [Left]  move window left
  [Right] move window right
  [Home]  update current file (re-read data from disk)
  [Space] cancel and deselect selection, return to pre-select position
          if playing, pause or continue playback
  <:      move cursor to sample 0
  >:      move cursor to last sample
  C-<:    move cursor to sample 0
  C->:    move cursor to last sample
  C-a:    move cursor to window start
  C-b:    move cursor back one sample
  C-d:    delete sample at cursor
  C-e:    move cursor to window end
  C-f:    move cursor ahead one sample
  C-g:    abort current command
  C-h:    delete previous sample
  C-i:    display cursor info
  C-j:    go to mark
  C-k:    delete one line's worth of samples
  C-l:    position window so cursor is in the middle
  C-m:    place (or remove) mark at cursor location
  C-n:    move cursor ahead one 'line'
  C-o:    insert one zero sample at cursor
  C-p:    move cursor back one 'line'
  C-q:    play current channel(s) starting at cursor
  C-r:    repeat last search backwards
  C-s:    search according to an expression
  C-t:    stop playing
  C-u:    start count (numerical argument) specification
  C-v:    move cursor to mid-window
  C-w:    delete current region
  C-x:    start extended command (see below)
  C-y:    paste in region
  C-z:    set current sample to 0.0
  C-_:    undo last edit
  C-[Space]: start region definition (arg<0 => cancel and deselect)

The extended commands (preceded by C-x) are:

  a:    apply envelope to selection
  b:    position window so cursor is on left margin
  c:    define selection from cursor to nth mark
  d:    get temp dir name
  e:    execute keyboard macro
  f:    position window so cursor is on right margin
  i:    insert region
  j:    go to named mark
  k:    close file
  l:    position selection in mid-view
  n:    re-evaluate expression over selection
  o:    move to next or previous graph
  p:    play selection or region n
  q:    mix in region
  r:    redo last undone edit
  u:    undo last edit
  v:    position window over current selection
  w:    save selection as file
  x:    evaluate expression over selection
  z:    smooth selection
  /:    place named mark
  (:    begin keyboard macro definition
  ):    end keyboard macro definition
                        
  C-a:  apply amplitude envelope
  C-b:  set x window bounds (preceded by first sample number)
  C-c:  hide ("close") controls
  C-d:  print
  C-f:  open file
  C-g:  abort command
  C-i:  insert file
  C-j:  go to mix console
  C-l:  load scm file
  C-m:  add named mark
  C-n:  re-evaluate expression
  C-o:  show ("open") controls
  C-p:  set (time domain display) window size (preceded by size)
  C-q:  mix in file
  C-r:  redo last undone edit
  C-s:  save file
  C-u:  undo last edit
  C-v:  set (time domain display) window size as percentage of total
  C-w:  save current channel in file
  C-x:  evaluate expression
 
*/

gint
Editor::button_press_handler (GdkEventButton *ev, TrackView *tv)

{
	if (current_trackview) {
		current_trackview->wavearea->set_state (GTK_STATE_NORMAL);
	} 

	current_trackview = tv;

	/* maybe merge these together */

	current_trackview->wavearea->grab_focus ();
	current_trackview->wavearea->set_state (GTK_STATE_ACTIVE);
	
	return FALSE;
}		

gint
Editor::button_release_handler (GdkEventButton *ev, TrackView *tv)

{
	if (!edits_legal) {
		return FALSE;
	}

	switch (ev->button) {
	case 3:
		track_edit_menu->popup (1, ev->time);
		return TRUE;
	}

	return FALSE;
}

gint
Editor::key_press_handler (GdkEventKey *ev, TrackView *tv)

{
	bool control;
	bool alt;
	bool meta;
	bool shift;
	gint ret;
	
	current_trackview = tv;

	control = (ev->state & GDK_CONTROL_MASK);
	shift = (ev->state & GDK_SHIFT_MASK);
	meta = (ev->state & GDK_MOD1_MASK);
	alt = (ev->state & GDK_MOD2_MASK);

	if (control_c_pending) {
		ret = do_control_c_keymap (ev, control, shift, meta, alt);
		control_c_pending = false;
	} else if (control_x_pending) {
		ret = do_control_x_keymap (ev, control, shift, meta, alt);
		control_x_pending = false;
	} else {
		if (meta || alt || escape_as_meta) {
			ret = do_meta_keymap (ev, control, shift, meta, alt);
			escape_as_meta = false;
		} else if (control) {
			ret = do_control_keymap (ev, control, shift, meta, alt);
		} else {
			ret = do_plain_keymap (ev, control, shift, meta, alt);
		}
	}

	/* Always stop the default handler from running, because it
	   will return FALSE, which will lead to our accelerators
	   running the same action again. Essentially, we don't
	   want to actually have the accelerators do *anything*.
	*/

	gtk_signal_emit_stop_by_name (GTK_OBJECT(tv->wavearea->gtkobj()), "key_press_event");

	return TRUE;
}

gint
Editor::do_control_x_keymap (GdkEventKey *ev, 
			     bool control, bool shift,
			     bool meta, bool alt)
{
	switch (ev->keyval) {
	case 'r':
		ARDOUR_UI::instance()->switch_to_notebook_page (0);
		break;
	case 'c':
                ARDOUR_UI::instance()->toggle_control_window ();
		break;
	case 'l':
		ARDOUR_UI::instance()->toggle_location_window ();
		break;
	case 'R':
		toggle_region_window ();
		break;
	}
	return TRUE;
}

gint
Editor::do_control_c_keymap (GdkEventKey *ev, 
			     bool control, bool shift,
			     bool meta, bool alt)
{
	switch (ev->keyval) {
	case 'b':
		cmd_indicator.set_text ("toggle bypass");
		/* bypass */
		break;

	case 'i':
		cmd_indicator.set_text ("toggle auto input");
		tape->set_auto_input (!tape->get_auto_input());
		break;

	case 'l':
		cmd_indicator.set_text ("toggle auto loop");
		tape->set_auto_loop (!tape->get_auto_loop());
		break;

	case 'p':
		cmd_indicator.set_text ("toggle auto play");
		tape->set_auto_play (!tape->get_auto_play());
		break;
	}

	return TRUE;
}

gint
Editor::do_plain_keymap (GdkEventKey *ev, 
			   bool control, bool shift,
			   bool meta, bool alt)
{
	switch (ev->keyval) {
	case GDK_Escape:
		escape_as_meta = true;
		break;

	case GDK_Home:
		move_to_start ();
		break;

	case GDK_End:
		move_to_end ();
		break;

	case GDK_Up:
		temporal_zoom_step (true);
		break;

	case GDK_Down:
		temporal_zoom_step (false);
		break;

	case GDK_Page_Up:
	case GDK_Left:
		scroll_back_screen ();
		break;

	case GDK_Page_Down:
	case GDK_Right:
		scroll_forward_screen ();
		break;

	case GDK_Insert:
		insert_cut_buffer ();
		break;

	case GDK_Delete:
		delete_sample_forward ();
		break;

	case GDK_equal:
		temporal_zoom_step (true);
		break;

	case GDK_minus:
		temporal_zoom_step (false);
		break;

	case GDK_less:
		move_to_start ();
		break;

	case GDK_greater:
		move_to_end ();
		break;

	case GDK_space:
		toggle_playback ();
		break;

	case GDK_r:
		reverse_selection ();
		break;

	case GDK_e:
		edit_envelope ();
		break;

	default:
		if (prefix_pending) {
			switch (ev->keyval) {
			case GDK_0:
				prefix_string += '0';
				break;
			case GDK_1:
				prefix_string += '1';
				break;
			case GDK_2:
				prefix_string += '2';
				break;
			case GDK_3:
				prefix_string += '3';
				break;
			case GDK_4:
				prefix_string += '4';
				break;
			case GDK_5:
				prefix_string += '5';
				break;
			case GDK_6:
				prefix_string += '6';
				break;
			case GDK_7:
				prefix_string += '7';
				break;
			case GDK_8:
				prefix_string += '8';
				break;
			case GDK_9:
				prefix_string += '9';
				break;
			case GDK_period:
				prefix_string += '.';
				break;
			}
		}
	}

	return TRUE;
}

gint
Editor::do_control_keymap (GdkEventKey *ev, 
			   bool control, bool shift,
			   bool meta, bool alt)
{
	switch (ev->keyval) {
	case GDK_a:
		move_to_start ();
		break;

	case GDK_b:
		scroll_back_sample ();
		break;

	case GDK_c:
		control_c_pending = true;
		break;

	case GDK_d:
		delete_sample_forward ();
		break;

	case GDK_e:
		move_to_end ();
		break;

	case GDK_f:
		scroll_forward_sample ();
		break;

	case GDK_h: /* backspace */
		delete_sample_backward ();
		break;

	case GDK_j:
		jump_to_mark ();
		break;

	case GDK_k:
		delete_screen ();
		break;

	case GDK_l:
		center_cursor ();
		break;

	case GDK_m:
		set_mark ();
		break;

	case GDK_n:
		scroll_forward_screen ();
		break;
		
	case GDK_o:
		insert_silence ();
		break;

	case GDK_p:
		scroll_back_screen ();
		break;

	case GDK_r:
		redo ();
		break;

	case GDK_s:
		silence_selection ();
		break;

	case GDK_t:
		// ??? 
		break;
		
	case GDK_u:
		prefix_pending = true;
		prefix_string = "";
		break;

	case GDK_v:
		scroll_tracks_down ();
		break;

	case GDK_w:
		cut_selection ();
		break;

	case GDK_x:
		control_x_pending = true;
		break;

	case GDK_y:
		insert_cut_buffer ();
		break;

	case GDK_z:
		undo ();
		break;

	case GDK_0:
		zero_sample_at_cursor ();
		break;

	case GDK_less:
		move_to_start ();
		break;

	case GDK_greater:
		move_to_end ();
		break;
	}

	return TRUE;
}

gint
Editor::do_meta_keymap (GdkEventKey *ev, 
			  bool control, bool shift,
			  bool meta, bool alt)
{
	switch (ev->keyval) {
	case GDK_v:
		scroll_tracks_up ();
		break;

	case GDK_w:
		cut_selection (false);
		break;
	}

	return TRUE;
}

/***********************************************************************
  Editor operations
 ***********************************************************************/

int
Editor::ensure_selection (gint32 *start, gint32 *cnt)

{
	char buf[64];

	if (!edits_legal) {
		return -1;
	}

	if (current_trackview == 0) {
		cmd_indicator.set_text ("No active track");
		return -1;
	}

	gtk_wave_view_get_selection (GTK_WAVE_VIEW (current_trackview->waveview), start, cnt);

	if (cnt == 0) {
		sprintf (buf, "Trk %d: no selection", current_trackview->tracknum);
		cmd_indicator.set_text (buf);
		return -1;
	}

	return 0;
}

int
Editor::ensure_cursor (gint32 *pos)

{
	char buf[64];

	if (!edits_legal) {
		return -1;
	}

	if (current_trackview == 0) {
		cmd_indicator.set_text ("No active track");
		return -1;
	}

	gtk_wave_view_get_cursor (GTK_WAVE_VIEW (current_trackview->waveview), pos);
	if (pos < 0) {
		sprintf (buf, "Trk %d: no cursor", current_trackview->tracknum);
		cmd_indicator.set_text (buf);
		return -1;
	}
	return 0;
}

/* DISPLAY MOTION */

void
Editor::move_to_start ()
{
	gdouble val;
	Gtk::Adjustment *adj;
	vector<TrackView *>::iterator i;

	if (current_trackview == 0) {
		return;
	}

	adj = current_trackview->x_adjuster;
	val = adj->get_lower ();

	if (current_trackview == region_view) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW(region_view->waveview), (int) adj->get_lower());
		return;
	}

	cmd_indicator.set_text ("move to start");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {

		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW((*i)->waveview), 
						     (int) (*i)->x_adjuster->get_lower());
	}

	tape->request_locate (0);
}

void
Editor::move_to_end ()
{
	gdouble val;
	Gtk::Adjustment *adj;
	vector<TrackView *>::iterator i;

	if (current_trackview == 0) {
		return;
	}

	adj = current_trackview->x_adjuster;
	val = adj->get_upper ();

	if (current_trackview == region_view) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW(region_view->waveview), (int) adj->get_lower());
		return;
	}

	cmd_indicator.set_text ("move to end");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW((*i)->waveview), 
						     (int) (*i)->x_adjuster->get_upper());
	}

	tape->request_locate (tape->frames());
}

void
Editor::center_cursor ()
{
	gdouble val;
	vector<TrackView *>::iterator i;
	gint32 pos;

	if (ensure_cursor (&pos)) {
		return;
	}
	
	val = (gdouble) pos - (current_trackview->x_adjuster->get_page_size()/2.0);

	if (current_trackview == region_view) {
		region_view->x_adjuster->set_value (val);
		return;
	}

	cmd_indicator.set_text ("center cursor");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		(*i)->x_adjuster->set_value (val);
	}
}

void
Editor::scroll_back_sample ()
{
	vector<TrackView *>::iterator i;
	gint32 pos;
	guint32 cnt;
	gdouble prefix;
	bool was_floating;
	char buf[64];

	if (ensure_cursor (&pos)) {
		return;
	}

	if ((prefix = use_prefix (&was_floating)) == 0) {
		sprintf (buf, "cursor back 1 sample");
		cnt = 1;
	} else {
		if (was_floating) {
			cnt = (guint32) floor (prefix * tape->frame_rate ());
			sprintf (buf, "cursor back %.3g second%s", prefix, prefix == 1.0 ? "" : "s");
		} else {
			cnt = (guint32) prefix;
			sprintf (buf, "cursor back %d sample%s", cnt, cnt == 1 ? "" : "s");
		}
	}

	if (current_trackview == region_view) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW(region_view->waveview), pos);
		return;
	}

	cmd_indicator.set_text (buf);

	if ((guint32) pos < cnt) {
		pos = 0;
	} else {
		pos -= cnt;
	}
	
	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW((*i)->waveview), pos);
	}

	/* XXX this is completely insane. with the current buffering
	   design, we'll force a complete track buffer flush and
	   reload, just to move 1 sample !!!
	*/

	tape->request_locate (pos);
}

void
Editor::scroll_forward_sample ()
{
	vector<TrackView *>::iterator i;
	gint32 pos;
	guint32 cnt;
	gdouble prefix;
	bool was_floating;
	char buf[64];

	if (ensure_cursor (&pos)) {
		return;
	}

	if ((prefix = use_prefix (&was_floating)) == 0) {
		sprintf (buf, "cursor forward 1 sample");
		cnt = 1;
	} else {
		if (was_floating) {
			cnt = (guint32) floor (prefix * tape->frame_rate ());
			sprintf (buf, "cursor forward %.3g second%s", prefix, prefix == 1.0 ? "" : "s");
		} else {
			cnt = (guint32) prefix;
			sprintf (buf, "cursor forward %d sample%s", cnt, cnt == 1 ? "" : "s");
		}
	}

	if (current_trackview == region_view) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW(region_view->waveview), pos+cnt);
		return;
	}

	cmd_indicator.set_text (buf);
	
	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_cursor_and_scroll (GTK_WAVE_VIEW((*i)->waveview), pos+cnt);
	}

	/* XXX this is completely insane. with the current buffering
	   design, we'll force a complete track buffer flush and
	   reload, just to move 1 sample !!!
	*/

	tape->request_locate (pos+cnt);
}

void
Editor::scroll_back_screen ()
{
	gdouble val;
	Gtk::Adjustment *adj;
	vector<TrackView *>::iterator i;

	if (current_trackview == 0) {
		return;
	}

	adj = current_trackview->x_adjuster;
	val = adj->get_value() - adj->get_page_size();

	if (current_trackview == region_view) {
		region_view->x_adjuster->set_value (val);
		return;
	}

	cmd_indicator.set_text ("scroll backward 1 page");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		(*i)->x_adjuster->set_value (val);
	}

	tape->request_locate ((size_t) current_trackview->x_adjuster->get_value());
}

void
Editor::scroll_forward_screen ()
{
	gdouble val;
	Gtk::Adjustment *adj;
	vector<TrackView *>::iterator i;

	if (current_trackview == 0) {
		return;
	}

	adj = current_trackview->x_adjuster;
	val = adj->get_value() + adj->get_page_size();

	if (current_trackview == region_view) {
		region_view->x_adjuster->set_value (val);
		return;
	}

	cmd_indicator.set_text ("scroll forward 1 page");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		(*i)->x_adjuster->set_value (val);
	}
	
	tape->request_locate ((size_t) current_trackview->x_adjuster->get_value());
}

void
Editor::scroll_tracks_down ()
{
	Gtk::Adjustment *adj = channel_scroller.get_hadjustment();
	adj->set_value (adj->get_value() + adj->get_page_size());
}

void
Editor::scroll_tracks_up ()
{
	Gtk::Adjustment *adj = channel_scroller.get_hadjustment();
	adj->set_value (adj->get_value() - adj->get_page_size());
}

/* ZOOM */

void
Editor::temporal_zoom (gdouble scale)

{
	vector<TrackView *>::iterator i;
	char buf[32];
	
	if (current_trackview == region_view) {
		gtk_wave_view_set_zoom (GTK_WAVE_VIEW(region_view->waveview), scale);
		return;
	}

	sprintf (buf, "set zoom to %.0g", scale);
	cmd_indicator.set_text (buf);

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_zoom (GTK_WAVE_VIEW((*i)->waveview), scale);
	}
}

void
Editor::temporal_zoom_step (bool in)

{
	vector<TrackView *>::iterator i;
	gdouble zoom;
	char buf[32];

	if (current_trackview == 0) {
		return;
	}

	zoom = gtk_wave_view_get_zoom (GTK_WAVE_VIEW(current_trackview->waveview));

	if (in) {
		zoom *= 2.0;
	} else {
		if (zoom > 2.0) {
			zoom /= 2.0;
		} else {
			zoom = 1.0;
		}
	}

	if (current_trackview == region_view) {
		gtk_wave_view_set_zoom (GTK_WAVE_VIEW(region_view->waveview), zoom);
		return;
	}

	sprintf (buf, "set zoom to %.0f", zoom);
	cmd_indicator.set_text (buf);

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_zoom (GTK_WAVE_VIEW((*i)->waveview), zoom);
	}
}	

void
Editor::temporal_zoom_selection ()

{
	vector<TrackView *>::iterator i;

	if (current_trackview == region_view) {
		gtk_wave_view_set_zoom_selection (GTK_WAVE_VIEW(region_view->waveview));
		return;
	}

	cmd_indicator.set_text ("zoom to selection");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_zoom_selection (GTK_WAVE_VIEW((*i)->waveview));
	}
}

void
Editor::temporal_zoom_tape ()

{
	vector<TrackView *>::iterator i;
	
	if (current_trackview == region_view) {
		gtk_wave_view_set_zoom_all (GTK_WAVE_VIEW(region_view->waveview));
		return;
	}

	cmd_indicator.set_text ("zoom to all");

	for (i  = visible_track_views.begin ();
	     i != visible_track_views.end ();
	     i++) {
		gtk_wave_view_set_zoom_all (GTK_WAVE_VIEW((*i)->waveview));
	}
}

/* DELETION */

Editor::Region *
Editor::save_region (TrackView *tv, guint32 start, guint32 cnt)

{
	sequence_t *edl = tape->track (tv->tracknum)->sequence();
	Region *region = new Region (edl, tv->tracknum);
	char buf[256];
	char start_time_buf[32];
	char end_time_buf[32];

	sample_to_smpte_clock (start, start_time_buf,
			       tape->frame_rate (), 
			       Config->smpte_frames_per_second);
	sample_to_smpte_clock (start+cnt, end_time_buf,
			       tape->frame_rate (), 
			       Config->smpte_frames_per_second);
			       
	sprintf (buf, "trk %d %s-%s", 
		 tv->tracknum + 1,
		 start_time_buf, end_time_buf);
	region->set_name (buf);

	edl = sequence_subsequence (edl, start, cnt);

	region->waveview_buffer = gtk_edl_wave_buffer_new (edl, tape->frame_rate ());

	regions.push_front (region);
	region_selector.rescan ();

	region_view->set_buffer (region->waveview_buffer);
	selected_region = region;
	gtk_wave_view_set_zoom_all (GTK_WAVE_VIEW(region_view->waveview));

	return region;
}

void
Editor::free_region (Region *region)

{
	regions.remove (region);
	delete region;

	/* XXX redisplay region selector */
}

void
Editor::cut_selection (bool actually_delete)

{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 start;
	gint32 cnt;
	Region *first_saved_region = 0;

	if (current_trackview == region_view) {
		return;
	}

	if (ensure_selection (&start, &cnt)) {
		return;
	}

	if (actually_delete) {
		sprintf (buf, "cut selection (%u-%u)", start, cnt);
	} else {
		sprintf (buf, "copy selection (%u-%u)", start, cnt);
	}
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {

		if ((*i) == current_trackview) {
			first_saved_region = save_region ((*i), start, cnt);
		} else {
			save_region ((*i), start, cnt);
		}
		
		if (actually_delete) {
			gtk_editable_wave_buffer_delete 
				(GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), start, cnt);
		}
	}

	delete valid_views;

	/* Now set the "cut buffer" to the first saved region. We do
	   it this way so that if we in fact cut from several tracks,
	   the "cut buffer" points to the region removed from the 
	   track that the edit was initiated on.
	*/
	
	cut_buffer = first_saved_region;
}

void
Editor::delete_sample_forward ()
{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;

	if (ensure_cursor (&pos)) {
		return;
	}

	sprintf (buf, "delete sample at %u", pos);
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_delete (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), pos, 1);
	}

	delete valid_views;
}

void
Editor::delete_sample_backward ()
{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;

	if (ensure_cursor (&pos)) {
		return;
	}
	
	sprintf (buf, "delete previous sample at %u", pos - 1);
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_delete (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), pos - 1, 1);
	}

	delete valid_views;
}

void
Editor::delete_screen ()
{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;
	guint32 cnt;

	if (ensure_cursor (&pos)) {
		return;
	}

	cnt = (guint32) current_trackview->x_adjuster->get_page_size();

	sprintf (buf, "delete chunk (%u-%u)", pos, pos+cnt);
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_delete (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), pos, cnt);
	}

	delete valid_views;
}

/* SEARCH */

void
Editor::search_backwards ()
{
	/* what ? */
}

void
Editor::search_forwards ()
{
	/* what ? */
}

/* MARKS */

void
Editor::jump_to_mark ()

{
	return;
}

void
Editor::set_mark ()

{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;
	gdouble prefix;
	bool was_floating;

	if (ensure_cursor (&pos)) {
		return;
	}

	if ((prefix = use_prefix (&was_floating)) == 0) {
		/* have cursor already */
	} else {
		pos = (gint32) prefix;
	}

	if (was_floating) {
		pos *= tape->frame_rate ();
		sprintf (buf, "insert mark at %.3g second%s", prefix, prefix == 1.0 ? "" : "s");
	} else {
		sprintf (buf, "set mark at sample %d", pos);
	}

	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_wave_view_set_marker (GTK_WAVE_VIEW((*i)->waveview), pos);
	}

	delete valid_views;
}

/* INSERT/REPLACE */

void
Editor::insert_silence ()
{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;
	guint32 cnt;
	gdouble prefix;
	bool was_floating;

	if (ensure_cursor (&pos)) {
		return;
	}

	if ((prefix = use_prefix (&was_floating)) == 0) {
		sprintf (buf, "insert 1 sample of silence");
		cnt = 1;
	} else {
		if (was_floating) {
			cnt = (guint32) floor (prefix * tape->frame_rate ());
			sprintf (buf, "insert %.3g second%s of silence", prefix, prefix == 1.0 ? "" : "s");
		} else {
			cnt = (guint32) prefix;
			sprintf (buf, "insert %d sample%s of silence", cnt, cnt == 1 ? "" : "s");
		}
	}

	cmd_indicator.set_text (buf);

	piece_t *piece = piece_new (silent_buffer, 0, cnt);
	
	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_insert (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), pos, piece);
	}

	delete valid_views;
}

void
Editor::insert_cut_buffer ()
{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;

	if (ensure_cursor (&pos)) {
		return;
	}

	if (cut_buffer == 0) {
		cmd_indicator.set_text ("Nothing in cut buffer");
		return;
	}

	sprintf (buf, "insert from cut buffer");
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_insert_sequence
			(GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), pos, cut_buffer->edl);
	}

	delete valid_views;
}

void
Editor::silence_selection ()

{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 start;
	gint32 cnt;
	piece_t *silence;

	if (ensure_selection (&start, &cnt)) {
		return;
	}

	sprintf (buf, "silence selection (%u-%u)", start, cnt);
	cmd_indicator.set_text (buf);

	silence = piece_new (silent_buffer, 0, cnt);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_replace (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), start, silence);
	}

	delete valid_views;
}

void
Editor::zero_sample_at_cursor ()

{
	vector<TrackView *>::iterator i;
	char buf[64];
	gint32 pos;
	guint32 cnt;
	gdouble prefix;
	piece_t *silence;
	bool was_floating;

	if (ensure_cursor (&pos)) {
		return;
	}

	if ((prefix = use_prefix (&was_floating)) == 0) {
		sprintf (buf, "silence 1 sample at %d", pos);
		cnt = 1;
	} else {
		if (was_floating) {
			cnt = (guint32) floor (prefix * tape->frame_rate ());
			sprintf (buf, "silence %.3g second%s at %d",
				 prefix, prefix == 1.0 ? "" : "s", pos);
		} else {
			cnt = (guint32) prefix;
			sprintf (buf, "silence %d sample%s at %d", 
				 cnt, cnt == 1 ? "" : "s", pos);
		}
	}
	cmd_indicator.set_text (buf);

	silence = piece_new (silent_buffer, 0, 1);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		gtk_editable_wave_buffer_replace (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), pos, silence);
	}

	delete valid_views;
}

/* UNDO/REDO */

void
Editor::undo ()

{
	vector<TrackView *>::iterator i;
	guint32 cnt = 0;
	bool was_floating;
	guint32 err = 0;
	char buf[32];

	if (!edits_legal) {
		return;
	}

	if (current_trackview == 0) {
		return;
	}

	if ((cnt = (int) floor (use_prefix (&was_floating))) == 0) {
		cnt = 1;
	}

	sprintf (buf, "undo %u", cnt);
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		if (gtk_editable_wave_buffer_undo
		    (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), cnt)) {
			err++;
		}
	}

	delete valid_views;

	if (err) {
		gdk_beep ();
	}
}

void
Editor::redo ()

{
	vector<TrackView *>::iterator i;
	guint32 cnt = 0;
	bool was_floating;
	guint32 err = 0;
	char buf[32];

	if (!edits_legal) {
		return;
	}

	if (current_trackview == 0) {
		return;
	}

	if ((cnt = (int) floor (use_prefix (&was_floating))) == 0) {
		cnt = 1;
	}

	sprintf (buf, "redo %u", cnt);
	cmd_indicator.set_text (buf);

	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		if (gtk_editable_wave_buffer_redo (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), cnt)) {
			err++;
		}
	}

	delete valid_views;

	if (err) {
		gdk_beep ();
	}
}

/* BUILT-IN EFFECTS */

void
Editor::reverse_selection ()

{

}

/* GAIN ENVELOPE EDITING */

void
Editor::edit_envelope ()

{
}

#if 0
void
Editor::normalize_selection ()

{
	gint32 start;
	gint32 cnt;
	piece_t *silence;

	if (ensure_selection (&start, &cnt)) {
		return;
	}
	
	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		if (gtk_editable_wave_buffer_normalize (GTK_EDITABLE_WAVE_BUFFER((*i)->waveview_buffer), start, cnt)) {
			err++;
		}
	}

	delete valid_views;
}

void
Editor::normalize_track ()

{
	gint32 start;
	gint32 cnt;
	piece_t *silence;

	
	
	vector<TrackView *> *valid_views = get_valid_views (current_trackview->edit_group);

	for (i  = valid_views->begin (); i != valid_views->end (); i++) {
		
		
	}

	delete valid_views;
}
#endif

/* PLAYBACK */

void
Editor::toggle_playback ()

{
	if (tape) {
		if (tape->transport_state() == Tape::Stopped) {
			cmd_indicator.set_text ("Start playback");
			tape->request_transport_state (Tape::Roll);
		} else {
			cmd_indicator.set_text ("Stop playback");
			tape->request_transport_state (Tape::Stopped);
		}
	}
}

gdouble
Editor::use_prefix (bool *was_floating)

{
	char *end;
	const char *start;
	gdouble val;

	if (prefix_string.length() == 0) {
		*was_floating = false;
		return 0;
	}

	if (prefix_string.find ('.') != string::npos) {
		*was_floating = true;
	} else {
		*was_floating = false;
	}

	start = prefix_string.c_str();
	val = strtod (start, &end);

	prefix_string = "";
	prefix_pending = false;

	if (start == end) {
		error << "bad prefix string \""
		      << prefix_string
		      << '"'
		      << endmsg;
		return 0;
	}

	return val;
}

void
Editor::search ()

{

}


void
Editor::active ()

{
	if (restart_scrolling_when_active) {
		start_scrolling ();
	}
}

void
Editor::inactive ()

{
	stop_scrolling ();
	restart_scrolling_when_active = true;
}

void
Editor::set_color (Editor::ColorType type)

{
	return;
}

/* SAVING */

int
Editor::save_tracknum (int tracknum)

{
	int ret;

	if ((ret = sequence_save (tape->track (tracknum)->sequence(), 0))) {
		error << "Editor: could not save track " 
		      << tracknum + 1
		      << endmsg;
	} 
	return ret;
}

void
Editor::save_track ()

{
	if (tape == 0 || current_trackview == 0) {
		return;
	}

	save_tracknum (current_trackview->tracknum);
}

void
Editor::save_track_as ()

{
	error << "save track as is not yet implemented" << endmsg;
}

int
Editor::save_all_tracks ()

{
	vector<TrackView *>::iterator i;

	for (i = visible_track_views.begin(); i != visible_track_views.end(); i++) {
		if (save_tracknum ((*i)->tracknum)) {
			return -1;
		}
	}
	return 0;
}

/* CURSOR SETTING AND MARKS AND STUFF */

void
Editor::cursor_set_handler (guint32 frame)

{
	if (tape) {
		tape->request_locate (frame);
	}
}

void
Editor::selection_started_handler (guint32 frame)

{

}

void
Editor::selection_finished_handler (GRange *range)

{
	
}

/* CLICK TRACKS */

int
Editor::build_click_buffers ()

{
	if (click_emphasis_sound != 0) {
		return 0;
	}

	if ((click_emphasis_sound = file_buffer_new (Config->click_emphasis_sound.c_str())) == 0) {
		error << "Editor: cannot load click emphasis sound from "
		      << Config->click_emphasis_sound
		      << endmsg;
		return -1;
	}
	
	if ((click_sound = file_buffer_new (Config->click_sound.c_str())) == 0) {
		error << "Editor: cannot load click emphasis sound from "
		      << Config->click_emphasis_sound
		      << endmsg;
		return -1;
	}

	click_emphasis_piece = 
		piece_new (click_emphasis_sound, 0, click_emphasis_sound->length (click_emphasis_sound));
	click_piece = piece_new (click_sound, 0, click_sound->length (click_sound));

	return 0;
}

void
Editor::print_click ()

{
	guint32 beat_cnt;
	guint32 sample_step;
	guint32 pos;
	sequence_t *edl;
	guint32 bpm;
	guint beats_per_bar;

	hide_click_window ();

	if (current_trackview == region_view) {
		/* XXX fix the track edit menu so that this is not even possible */
		return;
	}

	if (build_click_buffers()) {
		error << "Editor: no click sounds available."
		      << endmsg;
		return;
	}

	edl = tape->track (current_trackview->tracknum)->sequence();

	bpm = (guint32) click_bpm_adjustment.get_value ();
	beats_per_bar = (guint32) click_bpb_adjustment.get_value ();

	sample_step = (guint32) ((60.0/bpm) * tape->frame_rate());

	pos = 0;
	beat_cnt = 1;

	for (pos = 0; pos < tape->frames(); pos += sample_step) {
		piece_t *p;

		if (beat_cnt == 1) {
			p = click_emphasis_piece;
		} else {
			p = click_piece;
		}

		if (++beat_cnt > beats_per_bar) {
			beat_cnt = 1;
		}
		
		sequence_replace_piece (edl, pos, p);
	}

	gtk_editable_wave_buffer_set_modified (GTK_EDITABLE_WAVE_BUFFER(current_trackview->waveview_buffer), 
					       0, edl->length);
}

void
Editor::do_click ()

{
	if (click_window.is_visible()) {
		hide_click_window ();
	}
	show_click_window (current_trackview->tracknum);
}		     

void
Editor::build_click_window ()

{
	click_bpb_label.set_text ("Bar");
	click_bpb_vbox.pack_start (click_bpb_label, false, false);
	click_bpb_vbox.pack_start (click_bpb_spinner, false, false);

	click_note_label.set_text ("Note");
	click_note_vbox.pack_start (click_note_label, false, false);
	click_note_vbox.pack_start (click_note_spinner, false, false);

	click_bpm_scale.set_draw_value (true);
	click_bpm_scale.set_digits (0);
	click_bpm_scale.set_usize (200, -1);
	click_bpm_scale.set_name ("ClickSlider");

	click_upper_hpacker.set_spacing (5);
	click_upper_hpacker.pack_start (click_bpm_scale, true, true);
	click_upper_hpacker.pack_start (click_bpb_vbox, false, false);
	click_upper_hpacker.pack_start (click_note_vbox, false, false);

	click_lower_hpacker.set_spacing (10);
	click_lower_hpacker.set_border_width (10);
	click_lower_hpacker.pack_start (click_ok_button, true, false);
	click_lower_hpacker.pack_start (click_cancel_button, true, false);

	click_vpacker.pack_start (click_upper_hpacker, true, true);
	click_vpacker.pack_start (click_lower_hpacker, true, true);

	click_window.add (click_vpacker);
	click_window.set_name ("ClickWindow");
	click_window.set_modal (true);
	click_window.set_position (GTK_WIN_POS_MOUSE);

	click_ok_button.clicked.connect (slot (this, &Editor::print_click));
	click_cancel_button.clicked.connect (slot (this, &Editor::hide_click_window));
}	

void
Editor::show_click_window (gint tracknum)

{
	char buf[32];

	sprintf (buf, "Click in track %d", tracknum + 1);
	click_window.set_title (buf);
	click_window.show_all ();
}

void
Editor::hide_click_window ()

{
	click_window.hide_all ();
}
