/*
    Copyright (C) 1999 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: hammerfall.cc,v 1.14 2000/10/16 17:27:33 pbd Exp $
*/

#include <stdio.h>
#include <pbd/error.h>

#include "hammerfall.h"
#include "monitored_switch.h"
#include "alsa_device.h"

#include <pbd/stl_delete.h>

using namespace ALSA;

vector<MultiChannelHardware::ChannelDescription> *Hammerfall::stereo_channels = 0;

Hammerfall::Hammerfall (MultiChannelDevice *mcdp, snd_ctl_t *hndle)
	: MultiChannelHardware (),
	  handle (hndle)

{
	MonitoredSwitch *sw;

	mcd = mcdp;

	lock_status[0] = false;
	sync_status[0] = false;
	lock_status[1] = false;
	sync_status[1] = false;
	lock_status[2] = false;
	sync_status[2] = false;
	said_that_spdif_is_fine = false;

	sw = new MonitoredSwitch (handle, "ADAT1 Sync Check", 1000);
	switches.push_back (sw);
	sw->check_me.connect (slot (this, &Hammerfall::check_sync));

	sw = new MonitoredSwitch (handle, "ADAT2 Sync Check", 1000);
	switches.push_back (sw);
	sw->check_me.connect (slot (this, &Hammerfall::check_sync));

	sw = new MonitoredSwitch (handle, "ADAT3 Sync Check", 1000);
	switches.push_back (sw);
	sw->check_me.connect (slot (this, &Hammerfall::check_sync));

	set_input_monitor_mask (0);
}

Hammerfall::~Hammerfall ()

{
	vector_delete (&switches);
}

int 
Hammerfall::capabilities ()

{
	return  Cap_HardwareMonitoring|
		Cap_WordClock|
		Cap_ClockMaster|
		Cap_AutoSync|
		Cap_ClockLockReporting;
}

static void 
set_control_id (snd_control_t *ctl, const char *name)
{
	ctl->indirect = 0;
	strncpy ((char *) ctl->id.name, name, 43);
	ctl->id.name[43] = '\0';
	ctl->id.numid = 0;
	ctl->id.iface = SND_CONTROL_IFACE_PCM;
	ctl->id.device = 0;
	ctl->id.subdevice = 0;
	ctl->id.index = 0;
}

void
Hammerfall::check_sync (snd_ctl_t *handle, snd_control_t &ctl)

{
	if (strcmp ((const char *) ctl.id.name, "ADAT1 Sync Check") == 0) {
		check_sync_state (ctl.value.enumerated.item[0], 0);
	} else if (strcmp ((const char *) ctl.id.name, "ADAT2 Sync Check") == 0) {
		check_sync_state (ctl.value.enumerated.item[0], 1);
	} else if (strcmp ((const char *) ctl.id.name, "ADAT3 Sync Check") == 0) {
		check_sync_state (ctl.value.enumerated.item[0], 2);
	} else {
		error << "Hammerfall: unknown control \""
		      << ctl.id.name
		      << '"'
		      << endmsg;
	}
}

void
Hammerfall::check_sync_state (int val, int adat_id)

{
	bool lock;
	bool sync;

	/* S/PDIF channel is always locked and synced, but we only
	   need tell people once that this is true.

	   XXX - maybe nee to make sure that the rate matches our
	   idea of the current rate ?
	*/

	if (!said_that_spdif_is_fine) {
		char status;

		status = MultiChannelDevice::Lock|MultiChannelDevice::Sync;

		/* XXX broken! fix for hammerfall light ! */

		mcd->channel_status_change (24, status);
		mcd->channel_status_change (25, status);
		said_that_spdif_is_fine = true;
	}

	lock = (val & 0x1) ? true : false;
	sync = (val & 0x2) ? true : false;

	if (lock_status[adat_id] != lock ||
	    sync_status[adat_id] != sync) {
		broadcast_channel_status_change (lock, sync, 
						 adat_id*8,
						 (adat_id*8)+8);
	}

	lock_status[adat_id] = lock;
	sync_status[adat_id] = sync;
}

void
Hammerfall::broadcast_channel_status_change (bool lock, bool sync,
					     int lowchn, int highchn)

{
	int n;
	char status = 0;

	if (lock) {
		status |= MultiChannelDevice::Lock;
	} else {
		status |= MultiChannelDevice::NoLock;
	}

	if (sync) {
		status |= MultiChannelDevice::Sync;
	} else {
		status |= MultiChannelDevice::NoSync;
	}

	for (n = lowchn; n < highchn; n++) {
		mcd->channel_status_change (n, status);
	}
}

int
Hammerfall::set_input_monitor_mask (unsigned int mask)

{
      snd_control_t ctl;
      int err;
      int i;

      set_control_id (&ctl, "Channels Thru");

      for (i = 0; i < 26; i++) {
              ctl.value.integer.value[i] = (mask & (1<<i)) ? 1 : 0;
      }
              
      if ((err = snd_ctl_cwrite (handle, &ctl)) != 0) {
              error << "cannot set input monitoring ("
                    << snd_strerror (err)
                    << ')'
                    << endmsg;
	      return -1;
      }
      return 0;
}

int
Hammerfall::change_sample_clock (SampleClockMode mode)

{
 	snd_control_t ctl;
	int err;

	set_control_id (&ctl, "Sync Mode");

	switch (mode) {
	case AutoSync:
		ctl.value.enumerated.item[0] = 0;	
		break;
	case ClockMaster:
		ctl.value.enumerated.item[0] = 1;	
		break;
	case WordClock:
		ctl.value.enumerated.item[0] = 2;	
		break;
	}

	if ((err = snd_ctl_cwrite (handle, &ctl)) < 0) {
		error << "ALSA-Hammerfall: cannot set clock mode" 
		      << endmsg;
	}

	return 0;
}

vector<MultiChannelHardware::ChannelDescription> &
Hammerfall::stereo_descriptors ()

{
	ChannelDescription cd;

	if (stereo_channels != 0) {
		return *stereo_channels;
	}

	stereo_channels = new vector<ChannelDescription>;

	cd.name = "S/PDIF";
	cd.channels.push_back (24);
	cd.channels.push_back (25);

	stereo_channels->push_back (cd);

	cd.name = "ADAT 1+2";
	cd.channels.push_back (0);
	cd.channels.push_back (1);

	stereo_channels->push_back (cd);

	cd.name = "ADAT 3+4";
	cd.channels.push_back (2);
	cd.channels.push_back (3);

	stereo_channels->push_back (cd);

	cd.name = "ADAT 5+6";
	cd.channels.push_back (4);
	cd.channels.push_back (5);

	stereo_channels->push_back (cd);

	cd.name = "ADAT 7+8";
	cd.channels.push_back (6);
	cd.channels.push_back (7);

	stereo_channels->push_back (cd);

	return *stereo_channels;
}
