;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  soundcard.scm
;;;  simple class definition for soundcard config-variable type 
;;;
;;;  Bill Gribble <grib@billgribble.com> 2/21/2000
;;;  $Id: soundcard.scm,v 1.2 2000/10/04 16:45:31 pbd Exp $
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(depend 'simple-obj)
(depend 'enumeration)
(support 'soundcard)

;;  the "soundcard" class.  Soundcard definitions just live in the
;;  configuration file; the parameters of a defined soundcard are
;;  handed off to the C layer when you set a config-var that needs a
;;  soundcard.

(define <soundcard>
  (make-simple-class 'soundcard 
                     '(device mode type format rate nchannels latency)))

(define (soundcard? self)
  (eq? (simple-obj-type self) 'soundcard))

(define (make-soundcard . args)
  (let* ((kv-pairs    (map (lambda (elt) (cons (car elt) (cadr elt)))
                           (if (pair? args) (car args) '())))
         (file        (assq 'file: kv-pairs))
         (device      (assq 'device: kv-pairs))
         (mode        (assq 'mode: kv-pairs))
         (type        (assq 'type: kv-pairs))
         (format      (assq 'format: kv-pairs))
         (rate        (assq 'rate: kv-pairs))
         (nchannels   (assq 'nchannels: kv-pairs))
         (latency     (assq 'latency: kv-pairs)))
                          
    ;; the only sanity checking on device and file is that it's a string.
    (if (pair? device)
        (if (string? (cdr device))
            (set! device (cdr device))
            (error "make-soundcard: device spec must be a string."))
        (set! device #f))
    
    (if (pair? file)
        (if (string? (cdr file))
            (set! file (cdr file))
            (error "make-soundcard: file spec must be a string."))
        (set! file #f))
    
    ;; mode must be a string
    (if (pair? mode)
        (if (string? (cdr mode))
            (set! mode (cdr mode))
            (error "make-soundcard: mode must be a string."))
        (set! mode #f))
    
    ;; type must be a string
    (if (pair? type)
        (if (string? (cdr type))
            (set! type (cdr type))
            (error "make-soundcard: type must be a string."))
        (set! type #f))
    
    ;; and format must be a string
    (if (pair? format)
        (if (string? (cdr format))
            (set! format (cdr format))
            (error "make-soundcard: format must be a string."))
        (set! format #f))
    
    ;; rate and nchannels are integers, latency is a float (msecs) 
    (if (pair? rate)
        (if (integer? (cdr rate))
            (set! rate (cdr rate))
            (error "make-soundcard: 'rate' must be an integer"))
        (set! rate 44100))
    
    (if (pair? nchannels)
        (if (integer? (cdr nchannels))
            (set! nchannels (cdr nchannels))
            (error "make-soundcard: 'nchannels' must be an integer"))
        (set! nchannels 2))
    
    (if (pair? latency) 
        (if (number? (cdr latency))
            (set! latency (cdr latency))
            (error "make-soundcard: 'latency' must be a number"))
        (set! latency 10))
    
    (if (string? file) 
        (if device 
            (error "make-soundcard: specify only one of 'file' and 'device'")
            (set! device file)))

    (let ((self (make-simple-obj <soundcard>)))
      (soundcard:set-device! self device)
      (soundcard:set-mode! self mode)
      (soundcard:set-type! self type)
      (soundcard:set-format! self format)
      (soundcard:set-rate! self rate)
      (soundcard:set-nchannels! self nchannels)
      (soundcard:set-latency! self latency)
      self)))

(define (soundcard:get-device self)
  (simple-obj-getter self 'soundcard 'device))

(define (soundcard:set-device! self value)
  (if (string? value)
      (simple-obj-setter self <soundcard> 'device value)
      (error "soundcard:set-device!: device must be a string.")))


(define (soundcard:get-mode self)
  (simple-obj-getter self <soundcard> 'mode))

(define (soundcard:set-mode! self value)
  (if (string? value)
      (simple-obj-setter self <soundcard> 'mode value)
      (error "soundcard:set-device!: mode must be a string.")))


(define (soundcard:get-type self)
  (simple-obj-getter self <soundcard> 'type))

(define (soundcard:set-type! self value)
  (if (string? value)
      (simple-obj-setter self <soundcard> 'type value)
      (error "soundcard:set-device!: type must be a string.")))


(define (soundcard:get-format self)
  (simple-obj-getter self <soundcard> 'format))

(define (soundcard:set-format! self value)
  (if (string? value)
      (simple-obj-setter self <soundcard> 'format value)
      (error "soundcard:set-device!: format must be a string.")))


(define (soundcard:get-rate self)
  (simple-obj-getter self <soundcard> 'rate))

(define (soundcard:set-rate! self value)
  (if (integer? value)
      (simple-obj-setter self <soundcard> 'rate value)
      (error "soundcard:set-rate!: rate must be an integer.")))


(define (soundcard:get-nchannels self)
  (simple-obj-getter self <soundcard> 'nchannels))

(define (soundcard:set-nchannels! self value)
  (if (integer? value)
      (simple-obj-setter self <soundcard> 'nchannels value)
      (error "soundcard:set-nchannels!: nchannels must be an integer.")))


(define (soundcard:get-latency self)
  (simple-obj-getter self <soundcard> 'latency))

(define (soundcard:set-latency! self value)
  (if (number? value)
      (simple-obj-setter self <soundcard> 'latency value)
      (error "soundcard:set-latency!: nchannels must be a number.")))

