/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: tape.h,v 1.49 2000/11/04 04:53:33 pbd Exp $
*/

#ifndef __ardour_tape_h__
#define __ardour_tape_h__

#include <asm/atomic.h>
#include <sigc++/signal_system.h>

#include <string>
#include <queue>
#include <stack>

#include <pbd/thread.h>
#include <pbd/pool.h>
#include <pbd/error.h>
#include <pbd/lock_free_fifo.h>

#include <midi++/types.h>

#include "ardour.h"
#include "audiothing.h"
#include "configuration.h"
#include "location.h"

#include "seq.h"

namespace ALSA {
	class MultiChannelDevice;
};

namespace MIDI {
	class Port;
	class MachineControl;
};

namespace ARDOUR {

class Track;	

class Tape : public QMThread, public AudioThing

{
  public:
	enum RecordState {
		Disabled = 0,
		Enabled = 1,
		Recording = 2
	};

	enum State { 
		Stopped = 0,
		Roll,
		Forward,
		Rewind,
		Locating
	};
	
	struct Event {
	    enum Type {
		    TransportState,
		    PunchIn,
		    PunchOut,
		    AutoLoop
	    };

	    Type            type;
	    size_t          action_frame;
	    size_t          target_frame;
	    Tape::State     transport_state;
	    bool            remove;
	    Event          *next;

	    Event () {
		    remove = false;
		    next = 0;
		    action_frame = 0;
		    target_frame = 0;
	    }
	};

	enum StorageType {
		Interleaved,
		FilePerChannel
	};
		
	/* Constructor for a new tape */

	Tape (ALSA::MultiChannelDevice *,
	      MIDI::MachineControl *,
	      const string &name,
	      const string &path, 
	      size_t ntracks, 
	      size_t minutes,
	      size_t frames_per_second);

	/* Constructor for a tape required to already exist */

	Tape (ALSA::MultiChannelDevice *, 
	      MIDI::MachineControl *,
	      const string &name, 
	      const string &path);

	~Tape ();

	bool           ok() { return _ok; }
	const string &path()   { return _path; }

	Track *track (size_t n) { 
		if (n < _ntracks) {
			return tracks[n];
		} else {
			return 0;
		}
	}

	sequence_t **sequences();

	State transport_state () { return _transport_state; }

	size_t transport_frame () { 
		return _transport_frame;
	}

	bool get_record_enabled() { 
		return (record_status () >= Enabled);
	}

	RecordState record_status() {
		return (RecordState) atomic_read (&_record_status);
	}

	/* Record status signals */

	SigC::Signal0<void> RecordEnabled;
	SigC::Signal0<void> RecordDisabled;

	/* Transport mechanism signals */

	SigC::Signal0<void> TransportStateChange; /* generic */

	void request_transport_state (Tape::State);
	void request_locate (size_t frame);
	void goto_end ()   { request_locate (_frames);}
	void goto_start () { request_locate (0); }
	void use_rf_shuttle_speed () {
		current_shuttle_frames = non_shuttle_frames;
	}
	void set_shuttle_speed (size_t frames_per_roll);
	void set_rf_shuttle_speed (size_t frames_per_roll);
	bool tracks_still_loading ();

	int  wipe ();
	int wipe_track (size_t track_num, bool take_lock = true);

	int toggle_track_record_enable (size_t track);
	int toggle_track_monitor_input (size_t track);

	size_t frames ()    { return _frames; }
	size_t frame_rate() { return _frames_per_second; }
	size_t minutes ()   { return _frames / (_frames_per_second * 60); }
	size_t ntracks ()   { return _ntracks; }

	LocationStack *locations() { return &_locations; }

	void set_auto_punch_location (Location *loc);
	void remove_auto_punch_location (Location *loc);

	void set_auto_loop_location (Location *loc);
	void remove_auto_loop_location (Location *loc);

	void delete_auto_locations ();

	int save_locations ();
	int restore_locations ();

	int set_track_input_channel (size_t track, int channel);
	int set_track_output_channel (size_t track, int channel);
	int set_track_record_enable (size_t track, bool yn);
	
	void set_track_solo (size_t tracknum, bool yn);
	void set_track_mute (size_t tracknum, bool yn);

	bool set_xfade_length (size_t msecs);

	void maybe_enable_record ();
	void disable_record () {
		atomic_set (&_record_status, Disabled);
		RecordDisabled ();
	}


	/* toggles */

	void toggle_monitor_input (int);
	void toggle_record_enable (int);

	enum ControlType {
		AutoPlay,
		AutoLoop,
		Rehearse,
		AllSafe,
		DynamicPunch,
		RecordOneTouch,
		AutoInput,
		AutoPunch,
		SendMTC,
		MMCControl
	};

	SigC::Signal1<void,ControlType> ControlChanged;

	void set_auto_play (bool yn) { auto_play = yn; ControlChanged (AutoPlay); }
	void set_auto_loop (bool yn) { auto_loop = yn; ControlChanged (AutoLoop); }
	void set_rehearse (bool yn) { rehearse = yn; ControlChanged (Rehearse); }
	void set_all_safe (bool yn) { all_safe = yn; ControlChanged (AllSafe); }
	void set_dynamic_punch (bool yn) { dynamic_punch = yn; ControlChanged (DynamicPunch); }
	void set_record_one_touch (bool yn) { record_one_touch = yn; ControlChanged (RecordOneTouch); }
	void set_auto_input (bool yn);
	void set_auto_punch (bool yn) { auto_punch = yn; ControlChanged (AutoPunch); }
	void set_send_mtc (bool yn) { send_mtc = yn; ControlChanged (SendMTC); }
	void set_mmc_control (bool yn) { _mmc_control = yn; ControlChanged (MMCControl); }

	bool get_auto_play () { return auto_play; }
	bool get_auto_input () { return auto_input; }
	bool get_auto_loop () { return auto_loop; }
	bool get_auto_punch () { return auto_punch; }
	bool get_rehearse () { return rehearse; }
	bool get_all_safe () { return all_safe; }
	bool get_dynamic_punch () { return dynamic_punch; }
	bool get_record_one_touch () { return record_one_touch; }
	bool get_send_mtc () { return send_mtc; }
	bool get_mmc_control () { return _mmc_control; }

	void remove_event (size_t frame, Event::Type type);

	/* These are required by our AudioThing virtual
	   base class.
	*/
	
	int process (size_t nframes, size_t nbytes);
	int post_process ();
	unsigned int current_frame () { return _transport_frame; }
	size_t nchannels () { return _ntracks; }
	bool xrun_recoverable (ssize_t xrun_frames);
	void post_connect ();

  protected:
	friend class Track;
	bool waiting_for_tracks () { return atomic_read (&_waiting_for_tracks); }
	void wait_for_butler ();

  private:
	bool                     _ok;
	ALSA::MultiChannelDevice *mcd;
	MIDI::MachineControl     *mmc;
	MIDI::Port               *mtc_port;
	State                    _transport_state;
	atomic_t                 _record_status;
	size_t                   _transport_frame;
	size_t                    read_ahead_ready_mask;
	size_t                   _tracks_with_read_ahead;
	int                      _tracks_still_loading;
	atomic_t                 _waiting_for_tracks;
	string                   _path;
	size_t                   _frames;
	size_t                   _frames_per_second;
	size_t                   _ntracks;
	size_t                    capture_offset;
	size_t                    legal_recording_frame;
	unsigned int              tracks_continuing;
	bool                      stop_pending;

	pthread_mutex_t  load_lock;
	pthread_cond_t   loaded;

	/* toggles */

	bool auto_play;
	bool auto_punch;
	bool auto_loop;
	bool auto_looped;
	bool auto_input;
	bool rehearse;
	bool all_safe;
	bool record_one_touch;
	bool dynamic_punch;
	bool monitor_in;
	bool send_mtc;

	LockFreeFIFO<Event *> pending_events;

	/* this is held by the audio thread
	   whenever the tape is rolling, rewinding or
	   forwarding. it is released when it stops.
	*/

	pthread_mutex_t  transport_lock;
	bool             transport_lock_held;
	void             set_transport_state (Tape::State);
	void             stop_transport ();
	void             locate (size_t frame);

	void             set_last_roll_location (size_t);
	void             set_last_record_location (size_t);

	pthread_cond_t   butler_summons;
	pthread_cond_t   butler_paused;
	bool            _butler_summoned;
	atomic_t         butler_pause;
	void             summon_butler ();
	void             terminate_butler ();
	static void     *start_butler_thread (void *);

	Track **tracks;
	stack<size_t>    solo_stack;
	int              solo_track;

	size_t           current_shuttle_frames;
	size_t           non_shuttle_frames;

	LocationStack    _locations;

	SigC::Connection auto_punch_start_changed_connection;
	SigC::Connection auto_punch_end_changed_connection;
	void             auto_punch_start_changed (Location *);
	void             auto_punch_end_changed (Location *);

	SigC::Connection auto_loop_start_changed_connection;
	SigC::Connection auto_loop_end_changed_connection;
	void             auto_loop_start_changed (Location *);
	void             auto_loop_end_changed (Location *);

	/* Event list. Not handled with STL because
	   its too slow for our purposes, and because
	   the sort() method isn't supported for
	   anything except vector<>. 
	 */

	Event *          events;
	Event           *next_event;

	/* there can only ever be one of each of these */

	Event *auto_loop_event;
	Event *punch_out_event;
	Event *punch_in_event;

	/* events */

	void add_event (size_t action_frame, Event::Type type, size_t target_frame = 0);
	void merge_event (Event *);
	void _remove_event (size_t action_frame, Event::Type);

	/* reused MIDI quarter frame time code message */

	MIDI::byte mtc_msg[16];
	char       mtc_smpte_type;   /* encoding of SMTPE type for MTC */
	size_t     last_mtc_smpte_frame;

	void first_stage_init (ALSA::MultiChannelDevice *,
			       MIDI::MachineControl *,
			       const string &name,
			       const string &path);

	int second_stage_init (bool new_tracks);
	int check_tracks ();
	int read_header ();
	int write_header ();
	int really_load_tracks ();
	int load_tracks (bool new_tracks);
	int post_track_load ();
	int create ();
	void set_file_size ();

	void enable_record () {
		atomic_set (&_record_status, Recording);
		set_last_record_location (_transport_frame);
		RecordEnabled ();
	}
	
	void increment_transport_position (unsigned int val) {
		if (_frames - val < _transport_frame) {
			_transport_frame = _frames;
		} else {
			_transport_frame += val;
		}
	}

	void decrement_transport_position (unsigned int val) {
		if (val < _transport_frame) {
			_transport_frame -= val;
		} else {
			_transport_frame = 0;
		}
	}

	void set_transport_position (unsigned int val) {
		_transport_frame = val;
	}

 	void post_transport_motion ();
	static void *tape_loader_thread (void *arg);

	void *do_work();
	int interleaved_work ();
	int non_interleaved_work ();

	void set_next_event ();
	int  process_event (Event *);

	/* stuff to help the butler thread do efficient
	   read ahead sorting
	*/

	struct TrackSpace {
	    size_t index;
	    size_t space;
	};

	TrackSpace *track_spaces;

	static int track_space_compare (const void *ts1, const void *ts2);

	bool _mmc_control;
	void mmc_deferred_play (MIDI::MachineControl &);
	void mmc_stop (MIDI::MachineControl &);
	void mmc_record_strobe (MIDI::MachineControl &);
	void mmc_record_exit (MIDI::MachineControl &);
	void mmc_track_record_status (MIDI::MachineControl &, 
				      size_t track, bool enabled);
	void mmc_fast_forward (MIDI::MachineControl &);
	void mmc_rewind (MIDI::MachineControl &);
	void mmc_locate (MIDI::MachineControl &, const MIDI::byte *);
	void mmc_shuttle (MIDI::MachineControl &mmc, float speed, bool forw);
	void mmc_record_enable (MIDI::MachineControl &mmc, size_t track,
				bool enabled);

	bool mtc_enabled;
	int send_full_time_code ();
	int send_midi_time_code ();

	size_t adjust_apparent_position (size_t frames);
	
	void set_record_status ();
	
	/* These must be ONLY be called by the audio thread. No
	   run-time checking of this is done, however - I rely
	   on my own design skills ;)
	*/

	void audio_lock_transport ();
	void audio_unlock_transport ();

	bool actively_recording () {
		return record_status() == Recording;
	}
	
	int no_roll (size_t nframes, bool state_changing);
	
	size_t         target_locate_frame;
	Tape::State    target_transport_state;

	atomic_t _state_changing;
	bool state_changing() { return atomic_read (&_state_changing); }

	void         do_set_transport_state_in_another_thread 
                 		(Tape::State, size_t locate_frame = 0);
	static void *transport_thread (void *arg);

	void queue_event (Event *ev) {
		/* XXX handle FIFO overflow */
		pending_events.push (ev);
	}

	void do_channels_without_tracks ();

	int setup_interleaved_file ();
	int closedown_interleaved_file ();
	int fetch_interleaved_chunk ();
};

}; /* namespace ARDOUR */

#endif /* __ardour_tape_h__ */


