/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: track.h,v 1.41 2000/11/03 19:45:39 pbd Exp $
*/

#ifndef __ardour_track_h__
#define __ardour_track_h__

#include <asm/atomic.h>
#include <sigc++/signal_system.h>

#include <string>
#include <queue>

#include <pbd/thread.h>
#include <pbd/pool.h>

#include "ardour.h"
#include "configuration.h"
#include "tape.h"

#include "seq.h"

namespace ALSA {
	class MultiChannelDevice;
};

namespace ARDOUR {


class Track : public SigC::Object

{	
  public:
	Track (Tape &, 
	       ALSA::MultiChannelDevice *mcd,
	       size_t track_num,
	       size_t nbufs);

	~Track ();

	bool ok() { return _ok; }

	const string &tagname () { return _tagname; }
	const string &path() { return _path; }
	size_t current_take() { return _current_take; }

	/* this marks all in-memory copies of the track data as
	   invalid, and refills the buffers. because of thread
	   race conditions, it should normally be called
	   with `synchronous' == true. Certain internal
	   uses of this function do not need to do this, but
	   everybody else does.
	*/

	int invalidate_buffers (bool synchronous = true);

	int use_take (size_t take_num);
	int use_new_take ();
	void set_tagname (const string &str);

	int input_channel () const { return _input_channel; }
	int output_channel () const { return _output_channel; }
	bool record_enabled() { return atomic_read (&_record_enabled); }
	bool monitoring_input () { return _monitor_input; }

	void set_monitor_input (bool yn);
	void toggle_monitor_input () { 	set_monitor_input (!_monitor_input); }

	sequence_t *sequence() { return edl; }

	size_t delay() { return _delay; }
	
	float peak_power() { float x = _peak_power; _peak_power = 0; return x; }

	SigC::Signal0<void> take_changed;
	SigC::Signal0<void> tag_changed;

  protected:
	friend class Tape;

	int  roll (size_t nsamples, 
		   bool recording_enabled,
		   bool recording_is_legal,
		   bool &need_butler);

	int  no_roll (size_t nsamples, bool state_changing);

	void set_record_enable (bool yn);
	void toggle_record_enable () { set_record_enable (!record_enabled()); }
	void set_input_channel (int n);
	void set_output_channel (int n);
	void reset_monitor_input ();
	void set_delay (size_t n) { _delay = n; }

	bool set_mute (bool yn) { bool x = _muted; _muted = yn; return x; }

	int seek (size_t which_sample);

	int wipe ();

	void flush_dirty ();
	int  write_behind (bool &flush_required);
	int  read_ahead ();

  private:
	int                       fd;
	sequence_t               *edl;
	buffer_t                 *edl_buffer;
	Tape                     &tape;
	size_t                   _current_take;
	string                   _tagname;
	string                   _path;
	ALSA::MultiChannelDevice *mcd;
	size_t                    track_num;
	atomic_t                 _record_enabled;
	int                      _input_channel;
	int                      _output_channel;
	bool                     _ok;
	size_t                   _monitor_input;
	size_t                    file_sample;
	size_t                    playback_sample;
	size_t                   _delay;
	size_t                    delay_left;
	float                    _peak_power;
	int                       max_level;
	bool                     _muted;

	/* A private "implementation aid" */
	
	struct Buffer {
	    bool needs_refill;
	    bool needs_flush;
	    bool needs_async_flush;
	    size_t dirty;
	    size_t samples;
	    ARDOUR::Sample *buf;
	    size_t file_sample;

	    Buffer ();
	    ~Buffer ();

	    void set_size (size_t samples);
	};

	Buffer *playback_buffers;
	Buffer *capture_buffers;
	size_t nbufs;             /* needs to be 3 or more */
	size_t capture_buffer;    /* used by audio thread */
	size_t playback_buffer;   /* ditto */
	size_t refill_buffer;     /* used by butler thread */

	size_t xfade_sample;
	bool was_recording;
	bool was_rolling;

	size_t           capture_samples_left;
	ARDOUR::Sample  *capture_ptr;
	size_t           playback_samples_left;
	ARDOUR::Sample  *playback_ptr;

	int do_xfade (size_t offset, size_t samples, bool recording);
	int do_flush (size_t n);
	int do_refill (size_t n);
	int do_initial_fill ();

	int load_take (size_t take_num);
};


}; /* namespace ARDOUR */

#endif /* __ardour_track_h__ */










